
#include "trithorpe.h"


Expr::~Expr() { }

err Expr::beginPut() { return Str("bad assignment"); }
err Expr::finishPut() { return Str("bad assignment"); }
err Expr::emitCall() { return Str("bad function call"); }
bool Expr::isError() { return false; }
err Expr::getError() { return NIL; }
err Expr::emitAsABlock() { return piler->outputAsABlock(this); }


ConsExpr::ConsExpr(Compiler* c, Expr* _head, ConsExpr* _tail) : Expr(c), head(_head), tail(_tail) {
    what= tri('C','o','n','s');
}
ConsExpr::~ConsExpr() { delete head; delete tail; }
err ConsExpr::emitGet() { 
    tri e;
    if (tail) {
	e= tail->emitGet();    // recurse to emit backwards
	if (e != NIL) return e;
	out( tri('p','o','p') );  // remove all but the final (head) result
    }

    e= head->emitGet();
    return e;
}
err ConsExpr::emitGetEach() { 
    tri e;
    if (tail) {
	e= tail->emitGetEach();    // recurse to emit backwards
	if (e != NIL) return e;
    }

    e= head->emitGet();
    return e;
}
#ifdef DEBUG
void ConsExpr::print() {
    say << "ConsExpr(";
    if (tail) tail->print(); else say << " NULL_TAIL ";
    say << " :: ";
    head->print();
    say << ") ";
}
#endif
int ConsExpr::len() {
    ConsExpr* p= this;
    int z= 0;
    while ( p ) {
	++z;
	p= p->tail;
    }
    return z;
}


Builtin::Builtin(Compiler* c, tri _code, Expr* _a1, Expr* _a2, Expr* _a3, Expr* _a4, int _nresults, int _special )
    : Expr(c)
    , code(_code) 
    , a1(_a1) 
    , a2(_a2) 
    , a3(_a3) 
    , a4(_a4) 
    , nresults(_nresults) 
    , special(_special) 
{
    what= tri('B','u','i','l');
    MUST( nresults==0 || nresults==1 );
}
Builtin::~Builtin() { 
    if (a1) delete a1;
    if (a2) delete a2;
    if (a3) delete a3;
    if (a4) delete a4;
}
err Builtin::emitGet() {
    if (a3 && special==THIRD_FIRST ) a3->emitGet();
    if (a1) a1->emitGet();
    if (a2) a2->emitGet();
    if (a3 && special!=THIRD_FIRST ) a3->emitGet();
    if (a4) a4->emitGet();
    out( code );
    if (nresults<1) out( NIL ); // push a nil to make up for lack of a result
    return NIL;
}
#ifdef DEBUG
void Builtin::print() {
    say << "Builtin(" << code;
    if (a1) a1->print();
    if (a2) a2->print();
    if (a3) a3->print();
    if (a4) a4->print();
    say << ") ";
}
#endif


Verbatim::Verbatim(Compiler* c, int _numverbs, tri _verb1, tri _verb2) 
    : Expr(c)
    , numverbs(_numverbs) 
    , verb1(_verb1) 
    , verb2(_verb2) 
{
    what= tri('V','e','r','b');
}
Verbatim::~Verbatim() { }
err Verbatim::emitGet() { 
    out( verb1 );
    if (numverbs>1) out( verb2 );
    return NIL; 
}
#ifdef DEBUG
void Verbatim::print() {
    say << "Verbatim(" << verb1;
	if (numverbs>1) say << verb2;
    say << ") ";
}
#endif



Lit::Lit(Compiler* c, tri _lit) : Expr(c), lit(_lit) {
    what= tri('L','i','t');
}
Lit::~Lit() { }
err Lit::emitGet() {
    if ( ! lit.isInt() ) out( tri('i','m','m','/') );  // int does not need imm/ ; others do
    out( lit );
    return NIL;
}
#ifdef DEBUG
void Lit::print() { say << "Lit(" << lit << ") "; }
#endif



Error::Error(Compiler* c, err _err, Expr* _inner) : Lit(c, _err) , inner(_inner) {
    what= tri('E','r','r','o');
    Bad();
}
Error::~Error() { }
err Error::emitGet() {
    if (lit.isSym()) out( tri('i','m','m','/') );
    out( lit );
    return lit; // return the err!
}
bool Error::isError() { return true; }
err Error::getError() { return lit; }
#ifdef DEBUG
void Error::print() {
    say << "Error("; 
    if (inner) inner->print();
    say << " :: " << lit << ") ";
}
#endif



LocalVar::LocalVar(Compiler* c, int _index)
    : Expr(c)
    , index(_index)
{
    what= tri('L','v','a','r');
}

LocalVar::~LocalVar() { }
err LocalVar::beginPut() { /*nothing*/ return NIL; }
err LocalVar::finishPut() { out( tri('s','t','o','/') ); out(index); return NIL; }
err LocalVar::emitGet() { out( tri('r','c','l','/') ); out(index); return NIL; }
#ifdef DEBUG
void LocalVar::print() { say << "Var(" << index << ")" ; }
#endif



Slot::Slot(Compiler* c, Expr* _table, tri _key)
    : Expr(c)
    , table(_table)
    , key(_key)
{
    what= tri('S','l','o','t');
}

Slot::~Slot() {
    delete table;
}
err Slot::beginPut() {
    //table->emitGet();
    //out(tri('i','m','m','/')); out(key);
    return NIL;
}
err Slot::finishPut() {
    table->emitGet();
    out(tri('i','m','m','/')); out(key);
    out( tri('a','p','u','t') );
    return NIL;
}
err Slot::emitGet() { 
    table->emitGet();
    out(tri('i','m','m','/')); out(key);
    out( tri('a','g','e','t') );
    return NIL;
}
err Slot::emitCall() { 
    table->emitGet();
    out(tri('i','m','m','/')); out(key);
    return NIL;
}
#ifdef DEBUG
void Slot::print() { 
    say << "Slot{ ";
    table->print();
    say << " . " << key << " } ";
}
#endif



Assign::Assign(Compiler* c, Expr* _lhs, Expr* _rhs)
    : Expr(c)
    , lhs(_lhs)
    , rhs(_rhs)
{
    what= tri('A','s','s','i');
}

Assign::~Assign() {
    delete lhs;
    delete rhs;
}
err Assign::emitGet() { 
    //lhs->beginPut();
    rhs->emitGet();
    lhs->finishPut();
    //lhs->emitGet();   // would return the value assigned, but wrong -- may eval too much twice
    //out( tri(99999) );   // for now, just let assignment value be NIL..... or 99999
    return NIL;
}
#ifdef DEBUG
void Assign::print() { 
    say << "Slot{ ";
    lhs->print();
    say << " := ";
    rhs->print();
    say << " } ";
}
#endif



Binop::Binop(Compiler* c, tri _op, Expr* _one, Expr* _two)
    : Expr(c)
    , op(_op)
    , one(_one)
    , two(_two)
{
    what= tri('B','i','n','o');
}

Binop::~Binop() { delete one; delete two; }
err Binop::emitGet() { one->emitGet(); two->emitGet(); out(op); return NIL; }
#ifdef DEBUG
void Binop::print() { 
    say << "Binop(";
    one->print();
    say << " " << op << " ";
    two->print();
    say << ") ";
}
#endif

Call::Call(Compiler* c, Expr* _one, ConsExpr* _argList)
    : Expr(c)
    , one(_one)
    , argList(_argList)
{
    what= tri('C','a','l','l');
}

Call::~Call() { delete one; if (argList) delete argList; }
err Call::emitGet() {
    one->emitCall();
    if (argList) argList->emitGetEach();
    int n= argList ? argList->len(): 0;
    out( tri('c','a','l','/') );
    out( tri( 2 + n) );
    return NIL;
}
#ifdef DEBUG
void Call::print() { 
    say << "Call(";
    one->print();
    if (argList) argList->print(); else say<<" [] ";
    say << ") ";
}
#endif

IfExpr::IfExpr(Compiler* c, Expr* _cond, Expr* _thenBlock, Expr* _elseBlock)
    : Expr(c)
    , cond(_cond)
    , thenBlock(_thenBlock)
    , elseBlock(_elseBlock)
{
    what= tri('I','f','E','x');
}

IfExpr::~IfExpr() { delete cond; delete thenBlock; delete elseBlock; }

err IfExpr::emitGet() {

    cond->emitGet();
    out( tri('i','f') );

    if (thenBlock)
	thenBlock->emitAsABlock();
    else
	out( NIL );

    if (elseBlock)
	elseBlock->emitAsABlock();
    else
	out( NIL );

    return NIL;
}
#ifdef DEBUG
void IfExpr::print() { 
    IFD( say << "IfExpr("; )
    IFD( cond->print(); )
    if (thenBlock) thenBlock->print(); else say << "(no THEN)";
    if (elseBlock) elseBlock->print(); else say << "(no ELSE)";
    IFD( say << ") "; )
}
#endif

LoopExpr::LoopExpr(Compiler* c, Expr* _body)
    : Expr(c)
    , body(_body)
{
    what= tri('L','o','o','p');
}

LoopExpr::~LoopExpr() { delete body; }

err LoopExpr::emitGet() {

    out( tri('l','o','o','p') );
    body->emitAsABlock();
    
    return NIL;
}
#ifdef DEBUG
void LoopExpr::print() { 
    say << "LoopExpr(";
    body->print(),
    say << ") ";
}
#endif


/////////////////////////////////////////////////////////////////

int Compiler::addLocal( tri varName ) 
{
    int already= findLocal( varName );
    if ( already >= 0 ) return already;

    localVars.aAdd( varName ); 
    return localVars.aSize()-1;
}
int Compiler::findLocal( tri varName )
{
    if ( ! useLocal() ) return -1;

    int n= localVars.aSize();
    for (int i=0; i<n; i++) {
	if ( localVars.aGet(i) == varName ) return i;
    }
    return -1;
}


Compiler::Compiler( tri _src, tri _dest, tri _lobby )
    : src( _src )
    , isize( src.aSize() )
    , inext( 0 )
    , dest( _dest )
    , lobby( _lobby )
{
}


void Compiler::output(const tri& x) {
    CheckDeep( x );
    dest.aAdd(x);
    CheckDeep( dest );
}
void Expr::out(const tri& x) { piler->output(x); }

Expr* Compiler::parseSuffices( Expr* stem )
{
    Expr* e= NULL;
    ConsExpr* y= 0;

    while ( ! done() ) {
	tri tok= cur();

	switch (tok.peek()) {


	case SYM1('.'):
	{
	    advance();
	    if ( done() ) { e= new Error(this, "unexpected end after '.' in parseSuffices"); goto bail; }

	    tri key= cur();
	    advance();

	    stem= new Slot(this, stem, key);
	} break;

	case SYM1('('):
	{
	    advance();
	    if ( done() ) { e= new Error(this, "unexpected end after '(' in parseSuffices"); goto bail; }

	    Error* error= NULL;
	    y= parseExprList( &error );  // may be null, if no args

	    if ( error ) { e= new Error(this, "... in parseSuffices", error );  goto bail; }

	    MUST ( cur() == tri(')') );
	    advance();

	    stem= new Call(this, stem, y);
	    y= NULL;
	} break;

	default:
	    goto finish;
	}
    }
finish:
    return stem;
bail:
    if (stem) delete stem;
    if (y) delete y;
    return e;
}


Expr* Compiler::parseExpr( )
{
    Expr* y= 0;
    Expr* x= parseTerm();

    if (!x) return 0;

    while ( ! done() ) {
	tri tok= cur();

	switch (tok.peek()) {

	case SYM1('='):
	{
	    advance();
	    if ( done() ) goto bail;

	    y= parseExpr();
	    if (!y) goto bail;

	    x= new Assign(this, x, y);
	} break;


	case SYM1('+'):
	case SYM1('-'):
	case SYM1('*'):
	case SYM1('/'):
	case SYM1('%'):
	case SYM1('&'):
	case SYM1('|'):
	case SYM1('^'):
	case SYM2('=','='):
	case SYM2('!','='):
	case SYM2('<','='):
	case SYM2('>','='):
	case SYM1('<'):
	case SYM1('>'):
	{
	    advance();
	    if ( done() ) goto bail;

	    y= parseTerm();
	    if (!y) goto bail;

	    x= new Binop(this, tok, x, y);
	} break;

	default:
	    goto finish;
	}
    }
finish:
    return x;
bail:
    if (x) delete x;
    if (y) delete y;
    return 0;
}

Expr* Compiler::parseTerm()
{
    Expr* x= parseBasic();
    if (!x) return NULL;

    return parseSuffices(x);
}
Expr* Compiler::parseBasic()
{
	if ( done() )  return new Error(this, "unexpected END in parseBasic") ; // EOS

	tri tok= cur();
	advance();

IFD( if (verbose) say << "PT: tok=" << (int) tok.peek() << "\n"; ) 

	if ( tok.isInt() ) {
	    return new Lit(this, tok);
	} else if ( tok.isTab() ) {
	    return new Lit(this, tok);

	} else if ( tok == tri('g','e','t','u') ) {
	    return parseBuiltin( tri('g','e','t','u'), 1 );  // GetUp

	} else if ( tok == tri('p','u','t','u') ) {
	    return parseBuiltin( tri('p','u','t','u'), 2, 0 );  // PutUp

	} else if ( tok == tri('a','a','p','p') ) {		// aapp == aAdd()
	    return parseBuiltin( tri('a','a','p','p'), 2, 0 );

	} else if ( tok == tri('a','l','e','n') ) {
	    return parseBuiltin( tri('a','l','e','n'), 1 );

	} else if ( tok == tri('a','n','e','w') ) {
	    return parseBuiltin( tri('a','n','e','w'), 0 );

	} else if ( tok == tri('a','g','e','t') ) {
	    return parseBuiltin( tri('a','g','e','t'), 2 );

	} else if ( tok == tri('a','p','u','t') ) {
	    return parseBuiltin( tri('a','p','u','t'), 3, 0, Builtin::THIRD_FIRST );

	} else if ( tok == tri('a','n','e','w') ) {

	    return new Verbatim(this, 1, tri('a','n','e','w') );

	} else if ( tok == tri('f','u','n','c') ) {
	    return parseFun();
	} else if ( tok == tri('f','u','n') ) {
	    return parseFun();

	} else if ( tok == tri('i','f') ) {

	    return parseIf();

	} else if ( tok == tri('l','o','o','p') ) {

	    return parseLoop();

	} else if ( tok == tri('b','r','k') ) {
	    return new Verbatim( this, 2, tri(1), tri('b','r','k') );
	} else if ( tok == tri('b','r','e','a') ) {
	    return new Verbatim( this, 2, tri(1), tri('b','r','k') );

	} else if ( tok == tri('n','i','l') ) {
	    
	    return new Verbatim( this, 1, tri('n','i','l') );

	} else if ( tok == tri('H','o','m','e') ) {
	    
	    return new Verbatim( this, 1, tri('h','o','m','e') );


        } else if ( tok == tri('(') ) {

		// subexpr in parens
           Expr* z= parseExpr();
           EQ( tri(')'), cur() );
           advance();
           return z;

	} else if ( tok.isSym() ) {

	    // if it is followed by '=', make it a local
	    if ( useLocal() && cur() == tri('=') )  addLocal( tok );
		
	    int i= findLocal( tok );
	    if (0<=i) {
		return new LocalVar( this, i );
	    } else {
		Expr* LOBBY= new Verbatim( this, 1, tri('h','o','m','e') );
		return new Slot( this, LOBBY, tok );
	    }
	}

	return 0;
}

ConsExpr* Compiler::parseExprList( Error** errorOut )
{
    ConsExpr* z= NULL;

	while ( ! done()  &&  cur() != tri(')')  &&  cur() != tri('}') ) {

	    if ( cur() == tri(';') || cur() == tri(',') ) {
		advance();
		continue;
	    }

	    Expr* expr= parseExpr();
	    if (!expr) {
		if (z) delete z;
		*errorOut= new Error(this, Str("bad Expr in Expr List") );
		return NULL;
	    }
	    if (expr->isError()) {
		if (z) delete z;
		*errorOut= new Error(this, Str("... in Expr List"), expr );
		return NULL;
	    }

	    IFD( if (verbose) say << expr; )

	    z= new ConsExpr(this, expr, z);
	}

    // Empty block evaluates to NIL
    return z;
}


Expr* Compiler::parseBlock()
{

	if ( cur() != tri('{') )   return new Error(this, Str("`{' expected before Block") );
	advance();

	Error* error= NULL;
	Expr* z= parseExprList( &error );

	if ( error ) return new Error(this, Str("... in Block"),  error );

	if ( cur() != tri('}') )   {
	    if (z) delete z;
	    return new Error(this, Str("`}' expected after Block") );
	}
	advance();

	if ( !z ) z= new Verbatim(this, 1, NIL); // empty block evals to NIL

	return z;
}

Expr* Compiler::parseIf()
{
    Expr* c= parseExpr();
    if (c->isError()) { return new Error(this, "... in condition of if", c); }

    Expr* t= parseBlock();
    if (t->isError()) { delete c; return new Error(this, "... in then-branch of if", t); }

    Expr* e= NULL;

    if ( cur()== tri('e','l','s','e') ) {
	advance();

	e= parseBlock();
	if (e->isError()) { delete t; delete c; return new Error(this, "... in else-branch of if", e); }
    } else {
	// There is no else clause.
    }

    return new IfExpr( this, c, t, e );
}

Expr* Compiler::parseLoop()
{
    tri z;  z.setNewTab();

    Expr* body= parseBlock();
    if (body->isError()) { return new Error(this, "... in body of Loop", body); }
    
    return new LoopExpr( this, body );
}

err Compiler::outputAsABlock(Expr* expr)
{
    err e;
    {
    tri z("tcod");
CheckDeep(z); CheckDeep(dest);
    {
	tri saved_dest= dest;
	dest= z;
CheckDeep(z); CheckDeep(dest);
	e= expr->emitGet();
CheckDeep(z); CheckDeep(dest);
	dest= saved_dest;
CheckDeep(z); CheckDeep(dest);
    }

CheckDeep(z); CheckDeep(dest);
    if ( e == NIL ) { output( z ); } else { say << "ERROR IN outputAsABlock:" << e << "\n"; }
    }
  CheckDeep(dest);

    return e;
}

Expr* Compiler::parseFun()
{
    tri z;  z.setNewTab();

	if ( cur() != tri('(') )   return new Error(this, Str("`fun' not followed by `('") );
	advance();

	tri saveLocalVars= localVars;
	localVars.setNewTab();
	addLocal( tri('t','h','i','s') );
	addLocal( tri('t','h','i','m') );

	while ( cur() != tri(')') ) {

	    if ( cur() != tri(',') ) addLocal( cur() );
	    advance();
	}
	advance();

	z.setNewTab();
	z.aAdd( tri('t','c','o','d') );
	z.aAdd( tri('l','c','l','/') );   // local var info follows
	z.aAdd( tri() );  // to be fixed

	Expr* body= parseBlock();


	if (body) {
	    if (body->isError()) 
		return  new Error(this, Str("... in function definition"), body );

	    // Substitute our z for dest
	    tri saved_dest= dest;
	    dest= z;

	    body->emitGet();

	    dest= saved_dest;
	    delete body;

	} else {

	    z.aAdd( tri('n','i','l') );

	}

	z.aPut( 2, localVars );  // fix with num local args

	localVars= saveLocalVars;  // restore
    return new Lit(this, z);
}
    
Expr* Compiler::parseBuiltin(tri code, int nargs, int nresults, int special )
{
    Expr* z= NULL;
    Expr* a1= NULL;
    Expr* a2= NULL;
    Expr* a3= NULL;
    Expr* a4= NULL;

    if ( cur() != tri('(') ) { z= new Error(this, "builtin not followed by `(`"); goto end; }
    advance();

    if ( nargs>0 ) { a1= parseExpr(); if (a1->isError()) { z= new Error(this, "... in arg 1 to builtin", a1); a1=NULL; goto end; }}
    if ( cur() == tri(',') ) advance();
    if ( nargs>1 ) { a2= parseExpr(); if (a2->isError()) { z= new Error(this, "... in arg 2 to builtin", a2); a2=NULL; goto end; }}
    if ( cur() == tri(',') ) advance();
    if ( nargs>2 ) { a3= parseExpr(); if (a3->isError()) { z= new Error(this, "... in arg 3 to builtin", a3); a3=NULL; goto end; }}
    if ( cur() == tri(',') ) advance();
    if ( nargs>3 ) { a4= parseExpr(); if (a4->isError()) { z= new Error(this, "... in arg 4 to builtin", a4); a4=NULL; goto end; }}
    if ( cur() == tri(',') ) advance();

    if ( cur() != tri(')') ) { z= new Error(this, "builtin argument not followed by `)`"); goto end; }
    advance();
    
    z= new Builtin( this, code, a1, a2, a3, a4, nresults, special );
    a1= a2= a3= a4= 0;

end:
    if (a4) delete a4;
    if (a3) delete a3;
    if (a2) delete a2;
    if (a1) delete a1;
    return z;
}

tri CompileString(tri x, char* locals )
{
    tri z("tcod");
    Compiler* piler= new Compiler(x,z);
    piler->localVars.setNewTab();

    tri lcls(locals);
    for (int i=0; i<lcls.aSize(); i++) piler->addLocal( lcls.aGet(i) );

    Expr* ex= piler->parseExpr();

    if (!ex) return tri('b','a','d');
    if (! piler->done() ) return tri('m','o','r','e');

    ex->emitGet();
    delete ex;
    delete piler;
    return z;
}

