#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>


#include "trithorpe.h"

TriTerp* T;

void showStack() { T->dumpStack(); }

int 
trimain(int argc, const char* argv[])
{

#define BIG 999999

    char buf[ BIG ];
    char lib[ BIG ];

    int libfd= open( "lib.tri", O_RDONLY );
    if ( libfd < 0 ) { say << "\n*** CANNOT OPEN lib.tri\n"; Bomb("trimain: cannot open lib.tri"); }
    int libcc= read( libfd, lib, sizeof lib );
    if ( libcc < 0 ) { say << "\n*** CANNOT READ lib.tri\n"; Bomb("trimain: cannot read lib.tri"); }
    buf[libcc]= '\0';
    close( libfd );

    int cc= read( 0, buf, sizeof buf );
    if ( cc < 0 ) { say << "\n*** CANNOT READ stdin\n"; Bomb("trimain: cannot read stdin"); }
    buf[cc]= '\0';

    Error *error= NULL;
    TriTerp terp;
    T= &terp; // for gdb

    // Argc
    terp.home.aStore( tri('A','r','g','c'), tri(argc) );

    // argv
    for (int i=0; i<argc; i++) {
	terp.home.aPut( i, Str( argv[i] ) );
    }

    // Putc() hack
    terp.home.aStore( tri('P','u','t','c'), tri("tcod putc nil ret") );
    terp.home.aStore( tri('P','u','t','s'), tri("tcod puts nil ret") );
    terp.home.aStore( tri('A','b','o','r'), tri("tcod abor nil ret") );
    terp.home.aStore( tri('P','u','t','U'), tri("tcod getup nil ret") );
    terp.home.aStore( tri('G','e','t','U'), tri("tcod putup nil ret") );

    ////// first the lib.tri

    err e;
    tri libTokens= tri::NewCodVec(lib);
    tri libCode("tcod");
    Compiler libpiler( libTokens, libCode, terp.home );
    Expr* list= libpiler.parseExprList( &error );
    if ( error ) { say << "\n*** lib.tri ERROR " << error << "\n";  Bomb("trimain: compile lib.tri returns error"); }
    if (list) { e= list->emitGet(); delete list; }
    e= terp.testEval( libCode, NIL, NIL );
    if ( e != tri() ) { say << "\n*** lib.tri ERROR " << e << "\n";  Bomb("trimain: eval lib.tri returns error"); }

    /////// now the stdin program

    tri program= tri::NewCodVec(buf);
    tri code("tcod");

    char* VERBOSE =  getenv("VERBOSE");
    verbose= atoi( VERBOSE ? VERBOSE : "0" );

    Compiler piler(program, code, terp.home );
    
    list= piler.parseExprList( &error );
    if ( error ) { say << "\n*** ERROR " << error << "\n";  Bomb("trimain: eval stdin returns error"); }

    if (list) { e= list->emitGet(); delete list; }

    if (verbose) dumpVec( code, "stdin" );

    e= terp.testEval( code, NIL, NIL );
    if ( e != tri() ) { say << "\n*** ERROR " << e << "\n";  Bomb("trimain: eval stdin returns error"); }

    if (verbose) dumpVec( terp.home, "Home" );

    return 0;
}
